import asyncio
import json
import logging
import socket
import time
import traceback
from contextlib import asynccontextmanager, suppress
from pathlib import Path

import bleak

from . import build, common
from .bleio import BLEIOConnector
from .comm import Tasks, recv, send
from .storage import config

SHOULD_EXIT = False

conn_state = common.ConnectionState.DISCONNECTED
bleio: BLEIOConnector | None = None
scan_lock = asyncio.Lock()
running_progresses = {}


@asynccontextmanager
async def scan_locked():
    await scan_lock.acquire()
    yield
    scan_lock.release()


async def scan_for_devices():
    global bleio
    while not SHOULD_EXIT:
        async with scan_locked():
            logging.debug("Starting scan...")
            devices = await bleak.BleakScanner.discover(common.DEVICE_SEARCH_ROUND)
        scan_time = time.time()
        if bleio is not None and bleio.address in config.cached_devices:
            config.cache_device(
                bleio.address,
                config.cached_devices[bleio.address]["name"],
                scan_time,
            )
        for device in devices:
            if device.name is None or not device.name.startswith("GSG-"):
                continue
            config.cache_device(device.address, device.name[4:], scan_time)
        logging.debug("Finished scan.")
        await asyncio.sleep(common.DEVICE_SEARCH_PAUSE)
    logging.info("STOPPED SCAN")


async def stay_connected():
    global bleio, conn_state
    while not SHOULD_EXIT:
        if bleio is not None and not bleio._ble.is_connected:
            bleio = None
            conn_state = common.ConnectionState.DISCONNECTED

        if bleio is not None and bleio.address != config.target_device:
            with suppress(BaseException):
                conn_state = common.ConnectionState.DISCONNECTING
                await bleio.disconnect()
                conn_state = common.ConnectionState.DISCONNECTED
            bleio = None

        if bleio is None and config.target_device is not None and config.target_device in get_near_devices():
            async with scan_locked():
                conn_state = common.ConnectionState.CONNECTING
                bleio = BLEIOConnector(config.target_device)
                try:
                    await bleio.connect()
                    conn_state = common.ConnectionState.CONNECTED
                except Exception as e:
                    traceback.print_exception(e)
                    conn_state = common.ConnectionState.DISCONNECTED
                    bleio = None
        await asyncio.sleep(common.DEVICE_CONNECT_PAUSE)
    logging.info("STOPPED CONNECT")


def get_near_devices():
    return {
        address: device
        for address, device in config.cached_devices.items()
        if (device.get("last_seen") or 0) > time.time() - common.DEVICE_TOO_OLD
    }


async def handle_client(conn: socket.socket):
    global bleio
    packet = common.read_into(await recv(conn, 1), common.Querys)
    match packet:
        case common.Querys.GET_ALL_DEVICES:
            await send(
                conn,
                json.dumps(
                    config.cached_devices,
                ),
            )
        case common.Querys.GET_NEAR_DEVICES:
            await send(
                conn,
                json.dumps(get_near_devices()),
            )
        case common.Querys.GET_DEVICE_BY_ADDRESS:
            mac_address = (await recv(conn, 17)).decode("utf-8")
            await send(
                conn,
                json.dumps(config.cached_devices.get(mac_address)),
            )
        case common.Querys.GET_TARGET_DEVICE:
            if config.target_device is not None:
                await send(
                    conn,
                    json.dumps(
                        config.cached_devices.get(config.target_device)
                        or {
                            "name": "Unkown",
                            "last_seen": 0,
                            "address": config.target_device,
                        },
                    ),
                )
            else:
                await send(conn, "null")
        case common.Querys.SET_TARGET_DEVICE:
            mac_address = (await recv(conn, 17)).decode("utf-8")
            config.target_device = mac_address
            await send(conn, common.Success.OK)
        case common.Querys.UNSET_TARGET_DEVICE:
            config.target_device = None
            await send(conn, common.Success.OK)
        case common.Querys.KILL_DAEMON_PROCESS:
            global SHOULD_EXIT
            SHOULD_EXIT = True
            await send(conn, common.Success.OK)
        case common.Querys.GET_CONNECTION_STATE:
            if (
                conn_state == common.ConnectionState.CONNECTED
                and bleio is not None
                and bleio.address != config.target_device
            ):
                await send(conn, common.ConnectionState.INVALID)
            else:
                await send(conn, conn_state)
        case common.Querys.HUB_REBOOT:
            if bleio is not None:
                with suppress(BaseException):
                    await bleio.send_packet(b"&")
                await send(conn, common.Success.OK)
            else:
                await send(conn, common.Success.FAILED)
        case common.Querys.HUB_SYNC:
            dir_ = (await recv(conn, 1024)).decode("utf-8")

            if bleio is not None:
                await send(conn, common.Success.OK)
                tasks = Tasks(conn)
                await build.folder_sync(bleio, Path(dir_), tasks)
                await tasks.done()
            else:
                await send(conn, common.Success.FAILED)
        case common.Querys.HUB_START_PROGRAM:
            if bleio is not None:
                with suppress(BaseException):
                    await bleio.send_packet(b"P")
                await send(conn, common.Success.OK)
            else:
                await send(conn, common.Success.FAILED)
        case common.Querys.HUB_STOP_PROGRAM:
            if bleio is not None:
                with suppress(BaseException):
                    await bleio.send_packet(b"X")
                await send(conn, common.Success.OK)
            else:
                await send(conn, common.Success.FAILED)
        case _:
            await send(conn, common.Success.FAILED)

    conn.close()


async def main():
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    loop = asyncio.get_event_loop()
    logging.info(f"listening on {common.HOST}:{common.PORT}")
    server.bind((common.HOST, common.PORT))
    server.listen()
    server.setblocking(False)

    loop.create_task(stay_connected())
    loop.create_task(scan_for_devices())

    while not SHOULD_EXIT:
        conn, _ = await loop.sock_accept(server)
        loop.create_task(handle_client(conn))

    logging.info("STOPPED SERVER")

    async with scan_locked():
        if bleio is not None:
            await bleio.disconnect()
    logging.info("STOPPED CONNECTION")
    server.close()


def run_daemon():
    asyncio.run(main())


if __name__ == "__main__":
    run_daemon()



import asyncio
import base64
import hashlib
import json
import math
import os
import shutil
import subprocess
import tempfile
import time
import zlib
from pathlib import Path

import mpy_cross
import yaml

from .bleio import BLEIOConnector
from .comm import Task, Tasks


class ForceReconnect(BaseException): ...


mpy_cross.set_version("1.20", 6)


async def expect_OK(bleio: BLEIOConnector, ignore=b"="):
    while True:
        await asyncio.sleep(0.001)
        packet = bleio.get_packet()
        if not packet:
            continue
        nxt, _ = packet
        if nxt in ignore:
            continue
        if nxt != b"K":
            print(f"Expecting OK, Invalid response {nxt}, resetting connection")
            await bleio.send_packet(b"$")
            raise ForceReconnect
        return True


async def send_file(bleio: BLEIOConnector, file, task: Task | None = None):
    data = file.read()
    data = zlib.compress(data)
    data = base64.b64encode(data)
    if task is not None:
        await task.set_max(len(data))
    while True:
        await asyncio.sleep(0.001)
        chunk = data[:110]
        data = data[110:]
        if not chunk:
            break
        await bleio.send_packet(b"C", chunk)
        await expect_OK(bleio)
        if task is not None:
            await task.update(len(chunk))
    await bleio.send_packet(b"E")


def build_py(src: Path, dest: Path, src_dir: Path):
    result = subprocess.run(
        ["mpy-cross", "-c", "6.3", src.relative_to(src_dir), "-o", dest],
        check=False,
        cwd=src_dir,
        stderr=subprocess.PIPE,
    )
    if result.returncode != 0:
        print(result.stderr.decode("utf-8"))
        return False
    return True


def copy_py(src: Path, dest: Path, src_dir: Path):
    shutil.copy(src, dest)
    return True


def build_yaml(src: Path, dest: Path, src_dir: Path):
    with src.open() as f:
        data = yaml.safe_load(f)
    dest.write_text(json.dumps(data))
    return True


file_builders = {".py": (build_py, ".mpy"), ".yaml": (build_yaml, ".json")}


def build(src_dir: Path, build_dir: Path):
    for file in src_dir.glob("**"):
        if file.is_dir():
            if file == src_dir:
                continue
            folder = build_dir / file.relative_to(src_dir)
            if file.exists():
                os.makedirs(folder, exist_ok=True)
            else:
                os.rmdir(folder)
            continue
        # Skip stup files
        if file.suffix == ".pyi":
            continue
        builder, new_suffix = file_builders.get(file.suffix, (None, None, None))
        if not builder:
            print(f"Unable to build {file.suffix} file.")
            continue
        wanted_dest = build_dir / file.with_suffix(new_suffix).relative_to(src_dir)

        if not file.exists():
            os.remove(wanted_dest)
            continue

        result = builder(file, wanted_dest, src_dir)
        if not result:
            print(f"Failed to build {file.relative_to(src_dir).as_posix()}.")
            continue


async def sync_path(
    bleio: BLEIOConnector,
    file: Path,
    root_dir: Path,
    tasks: Tasks,
):
    path = file.relative_to(root_dir).as_posix()
    if path == ".":
        return

    if not file.exists():
        await bleio.send_packet(b"R", ("/" + path).encode())
    elif file.is_dir():
        await bleio.send_packet(b"D", ("/" + path).encode())
        await expect_OK(bleio)
    else:
        hashv = hashlib.sha256(file.read_bytes()).hexdigest()
        await bleio.send_packet(b"F", ("/" + path + " " + hashv).encode())
        while True:
            await asyncio.sleep(0.001)
            packet = bleio.get_packet()
            if packet is None:
                continue
            nxt, _ = packet
            if nxt == b"=":
                continue
            if nxt == b"K":
                break
            if nxt != b"U":
                print(
                    f"Expecting OK or U, Invalid response {nxt}, resetting connection",
                )
                await bleio.send_packet(b"$")
                raise ForceReconnect

            max_ = math.ceil(file.stat().st_size * (4 / 3))
            async with tasks.task(f"Sync file {path}...", max_, "B") as task:
                with file.open("rb") as f:
                    await send_file(bleio, f, task)


async def sync_dir(
    bleio: BLEIOConnector,
    dir: Path,
    tasks: Tasks,
):
    await bleio.send_packet(b"Y")
    await expect_OK(bleio)

    files = tuple(dir.glob("**"))
    max_prog = len(files)
    async with tasks.task("Sync directory...", max_prog, "files") as task:
        for file in files:
            await sync_path(bleio, file, dir, tasks)
            await task.update(1)

    await bleio.send_packet(b"N")
    await expect_OK(bleio)


async def sync_stream(bleio: BLEIOConnector, timeout=10):
    otm = time.time()
    otm_packet = str(otm).encode()
    await bleio.send_packet(b"=", otm_packet)
    while time.time() < otm + timeout:
        await asyncio.sleep(0.001)
        resp = bleio.get_packet()
        if resp is None:
            await bleio.send_packet(b"=", otm_packet)
            await asyncio.sleep(1)
            continue
        nxt, tm = resp
        if nxt == b"=" and tm == otm_packet:
            return
    raise TimeoutError(f"Hub failed to sync after {timeout} seconds.")


async def folder_sync(
    bleio: BLEIOConnector,
    src_dir: Path,
    tasks: Tasks,
):
    with tempfile.TemporaryDirectory() as BUILD_DIR:
        await sync_stream(bleio, 10)

        build(src_dir, Path(BUILD_DIR))

        await sync_dir(bleio, Path(BUILD_DIR), tasks)

        await bleio.send_packet(b"P")
        await expect_OK(bleio)
